// v 3.3 :  gamma fit suppressed 1st iteration if initial gammas
//          from data set

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include <dos.h>
#include <dir.h>
#include <alloc.h>
#include <ctype.h>
#include <string.h>
#include <mem.h>

#include "standard.h"

// default is 4-byte floating point (REAL*4)
// ..change following to double for 8-byte floating point

#define FPPREC float

char *version="GOM 3.4 (Aug 92)";

int MAXITER=500;
int MAXITER2=25;

#define MAXVARS 100
#define MISSING -1
#define FMISSING ( FPPREC ) MISSING
#define MAXRECLEN 255
#define TOKENLEN 80
#define MAXSUBJECTS 4000
#define MAXTOKENS 64

char tempstring[255];

File *datafile=NULL;
File *tempfile = NULL;
File *inlambdafile = NULL;
File *ingammafile = NULL;
File *controlfile = NULL;
File *logfile = NULL;
File *printfile = NULL;
File *cellfile = NULL;
File *lamfile = NULL;
File *gamfile = NULL;

char *datafilename=NULL;
char *inlambdafilename=NULL;
char *ingammafilename=NULL;
char *controlfilename=NULL;
char *printfilename=NULL;
char *rootfilename=NULL;

char *logfilename=NULL;
char *cellfilename=NULL;
char *gamfilename=NULL;
char *lamfilename=NULL;

char *model_name=NULL;

typedef struct {
   int x;
   int y;
   int attr;
   int left;
   int top;
   int right;
   int bottom;
   int dim;
   int bright;
   int reverse;
   int hilite;
} WIND;

#define LOGWIN 0
#define MSGWIN 1
#define QWIN 2

int CURWIN=0;
int NWIN=3;

WIND Window[10];

/*
   screen layout
*/
int n_logfile     = 0;

void initwind(int W,int left,int top,int right,int bottom,
   int dim,int bright,int reverse,int hilite) {

   Window[W].left       = left    ;
   Window[W].top        = top     ;
   Window[W].right      = right   ;
   Window[W].bottom     = bottom  ;
   Window[W].dim        = dim     ;
   Window[W].bright     = bright  ;
   Window[W].reverse    = reverse ;
   Window[W].hilite     = hilite  ;
   Window[W].x          = 1       ;
   Window[W].y          = 1       ;
   Window[W].attr       = bright  ;
   CURWIN=W;
}


void window_box(int W,char *s) {
   int i;
   window(1,1,80,25);
   textattr(Window[W].hilite);

   gotoxy(Window[W].left ,Window[W].top);       putch('');
   gotoxy(Window[W].left ,Window[W].bottom);    putch('');
   gotoxy(Window[W].right,Window[W].top);      putch('');
   gotoxy(Window[W].right,Window[W].bottom);   putch('');

   for (i=Window[W].left+1;i<Window[W].right;i++) {
      gotoxy(i,Window[W].top);    putch('');
      gotoxy(i,Window[W].bottom); putch('');
   }

   for (i=Window[W].top+1;i<Window[W].bottom;i++) {
      gotoxy(Window[W].left,i);   putch('');
      gotoxy(Window[W].right,i);  putch('');
   }

   gotoxy(Window[W].left+2,Window[W].top);     cputs(s);

   window(Window[W].left+1,Window[W].top+1,
      Window[W].right-1,Window[W].bottom-1);

   textattr(Window[W].dim);

   clrscr();
}

void fullscreen();
void activate_window(int w);
void save_screen();
void restore_screen();

void diddle() {
   static int DIDDLE=0;
   fullscreen();
   textattr(WHITE);
   gotoxy(80,1);
   if (DIDDLE)
      putch('');
   else
      putch('');
   activate_window(CURWIN);
   DIDDLE=1-DIDDLE;
}

int query(char *s) {
   int w,ch;
   w=CURWIN;
   save_screen();
   window_box(QWIN,"Question");
   say(Window[QWIN].bright,2,3,s);
   cursor_on();
   beep();
   ch=getch();
   restore_screen();
   activate_window(w);
   if (ch=='y' || ch=='Y')
      return 1;
   else
      return 0;
}

int CANCELLED=0;

int cancelled() {
   int can;
   if (CANCELLED)
      return 1;
   can=0;
   while ( kbhit() )
      if (getch()==27)
         can=1;
   if (can) {
      if (query("Ok to cancel (y/n) ? ")) {
         CANCELLED=1;
      } else
         can=0;
   }
   return can;
}

/*
   input data structures
*/
char *varname[MAXVARS+1];
char internal[MAXVARS+1];
char external[MAXVARS+1];
char srcvarpos[MAXVARS+1];

char **subjid;
char **subjrec;
char **cell;
int  *cellnum;
int maxlevels=0;

/*
   model parameters
*/
FPPREC **G;

FPPREC ***P;
FPPREC ***new_P;
FPPREC ***old_P;
FPPREC ***dL_dP;
char   ***free_P;
int    **fit_P;
/*
   computational data structures
*/
FPPREC N;
FPPREC *cellfrq;

int I;

/*
   model specification
*/
int K=0;    /* #profiles      */
int J=0;    /* #internal vars */
int JX=0;   /* #external vars */
int TJ=0;   /* #vars (=J+JX)  */

/*
   algorithm controls
*/

#define REALSMALL 1.0E-10
#define REALBIG 1.0E+30

#define ZERO 1.0E-20

// convergence tolerance
#define CTOL 0.00001
#define ITOL 0.0005

#define WOODBURY74 0
#define GRADIENT 1
#define ZHANG_GRADIENT 2
#define ZHANG_EM 3

#define NALGS 4

char algname[NALGS][80]={
   "Woodbury 1974",
   "Gradient",
   "Zhang gradient algorithm for dichotomous variates",
   "Zhang EM algorithm for dichotomous variates"
};

#define DEFAULT 0
#define DATASET 1
#define RANDOM 2
#define PURE1 3

int DICHOTOMOUS=0;
int algorithm=WOODBURY74;
int G_algorithm=WOODBURY74;
int P_algorithm=WOODBURY74;
int iter=0;
int subiter=0;
int NOID=0;
int INITLAMBDA=DEFAULT;
int INITGAMMA=DEFAULT;
int FITLAMBDA=1;
int FITGAMMA=1;
int Gradient=1;
int P_gradient=0;
int G_gradient=0;

char INITSRCLABEL[4][80]={
   "Default (equal values)",
   "From dataset",
   "Random",
   "Pure Type I"
};

FPPREC loglik, ploglik, dloglik;

int keylen=10; /* max record key length */

int nsubjects=0;

int *levels;
int *baselevel;
FPPREC **freqtable;

int nvars=0;
int ERRORS=0;
int WARNINGS=0;

void crash(char *s) {
   fullscreen();
   textattr(WHITE);
   clrscr();
   printf("\n\nGOM crashed for the following reason:\n\n\t%s\n\n",s);
   Exit(1);
}

void fullscreen() {
   window(1,1,80,25);
}

void activate_window(int W) {
   cursor_off();
   window(Window[W].left+1,Window[W].top+1,
      Window[W].right-1,Window[W].bottom-1);
   gotoxy(Window[W].x,Window[W].y);
   textattr(Window[W].bright);
   CURWIN=W;
}

char *screenbuf;

void save_screen() {
   screenbuf=(char *)nulmemory(4000);
   fullscreen();
   gettext(1,1,80,25,screenbuf);
   activate_window(CURWIN);
}

void restore_screen() {
   fullscreen();
   puttext(1,1,80,25,screenbuf);
   activate_window(CURWIN);
   free(screenbuf);
}

void logwin() {
   activate_window(LOGWIN);
}

void msgwin() {
   activate_window(MSGWIN);
}

void draw_mainscreen() {

   initwind(LOGWIN,1,3,80,16,
      COLORATTR( LIGHTGRAY, BLACK ),
      COLORATTR( WHITE,     BLACK ),
      COLORATTR( BLACK, LIGHTGRAY ),
      COLORATTR( YELLOW,    BLACK ));

   initwind(MSGWIN,1,17,80,22,
      COLORATTR( BLACK,     CYAN  ),
      COLORATTR( WHITE,     CYAN  ),
      COLORATTR( YELLOW,    BLUE  ),
      COLORATTR( BLUE,      CYAN  ));

   initwind(QWIN,10,9,70,13,
      COLORATTR( LIGHTGRAY, RED   ),
      COLORATTR( WHITE,     RED   ),
      COLORATTR( BLACK, LIGHTGRAY ),
      COLORATTR( YELLOW,    RED   ));

   fullscreen();

   clrscr();

   say(WHITE,1,1,version);

   sprintf(tempstring,"Log File (%s)",logfilename);
   window_box(LOGWIN,tempstring);
   window_box(MSGWIN,"Progress");

   msgwin();
}

void log_message(char *s) {
   logwin();
   cprintf("\r%s\n\r",s);
   Window[LOGWIN].x=wherex();
   Window[LOGWIN].y=wherey();
   n_logfile++;
   Fwrite(logfile,s);
   fflush(logfile->stream);
}

void message(char *s) {
   msgwin();
   cprintf("\r%s\n\r",s);
   Window[MSGWIN].x=wherex();
   Window[MSGWIN].y=wherey();
}

void fatal_error(char *s) {
   sprintf(tempstring,"Error: %s",s);
   log_message(tempstring);
   log_message("");
   ERRORS++;
}

void warn(char *s) {
   sprintf(tempstring,"Warning: %s",s);
   log_message(tempstring);
   log_message("");
   WARNINGS++;
}

void howmuchleft() {
   sprintf(tempstring,"%lu bytes free.",farcoreleft());
   message(tempstring);
}

void filename(char **t,char *s) {
   if (*t == NULL)
      *t=(char *)getmemory(MAXPATH);
   memset(*t,0,MAXPATH);
   strncpy(*t,s,MAXPATH-1);
   unquote(*t);
}

void validate_specs() {
   int i;

   if (!nvars)
      fatal_error("INPUT statement missing or invalid.");

   if (!J)
      fatal_error("INTERNAL statement missing or invalid.");

   if (!datafilename)
      fatal_error("INFILE statement missing or invalid.");

   if (!K)
      fatal_error("PROFILES statement missing or invalid.");

   if (J && JX)
      for (i=0;i<nvars;i++)
         if (internal[i] && external[i]) {
            sprintf(tempstring,"Variable %s defined as both INTERNAL and EXTERNAL.",
               varname[i]);
            fatal_error(tempstring);
         }

   TJ = J + JX;

   if (!nsubjects) {
      nsubjects=MAXSUBJECTS+1;
      log_message("Default maximum #subjects assumed.");
   }

}

void open_log(char *ctrlfname) {
   int i;
   if (logfilename==NULL) {
      logfilename=(char *)getmemory(MAXPATH);
      memset(logfilename,0,MAXPATH);
      for (i=0;i<strlen(ctrlfname);i++)
         if (ctrlfname[i]=='.')
            break;
         else
            logfilename[i]=tolower(ctrlfname[i]);
      strcat(logfilename,".log");
   }
   logfile=Fopen(logfilename,"wt");
   if (logfile==NULL) {
      sprintf(tempstring,"Could not open LOG file %s",logfilename);
      crash(tempstring);
   }
}

void open_model_files(char *s) {
   char *drive, *name, *ext, *dir;
   drive=(char *)nulmemory(MAXDRIVE);
   name=(char *)nulmemory(MAXFILE);
   ext=(char *)nulmemory(MAXEXT);
   dir=(char *)nulmemory(MAXDIR);
   fnsplit(rootfilename,drive,dir,name,ext);
   if (model_name==NULL)
      model_name=(char *)nulmemory(9);
   if (s==NULL)
      strcpy(model_name,name);
   else
      strncpy(model_name,s,8);
   if (cellfilename==NULL)
      cellfilename=(char *)nulmemory(MAXPATH);
   if (gamfilename==NULL)
      gamfilename=(char *)nulmemory(MAXPATH);
   if (lamfilename==NULL)
      lamfilename=(char *)nulmemory(MAXPATH);
   fnmerge(cellfilename,drive,dir,model_name,".cel");
   fnmerge(gamfilename,drive,dir,model_name,".gam");
   fnmerge(lamfilename,drive,dir,model_name,".lam");
   cellfile=Fopen(cellfilename,"wt");
   gamfile=Fopen(gamfilename,"wt");
   lamfile=Fopen(lamfilename,"wt");
   if (cellfile==NULL || gamfile==NULL || lamfile==NULL)
      crash("Error opening output model files.");
   free(drive);
   free(dir);
   free(name);
   free(ext);
}

void open_files() {

   if (datafilename==NULL)
      crash("No input data file was specified.");

   datafile=Fopen(datafilename,"rt");
   if (datafile==NULL)
      crash("Error opening input data file.");

   if (inlambdafilename != NULL) {
      inlambdafile=Fopen(inlambdafilename,"rt");
      if (inlambdafile==NULL)
         crash("Error opening input lambda file.");
   }

   if (ingammafilename != NULL) {
      ingammafile=Fopen(ingammafilename,"rt");
      if (ingammafile==NULL)
         crash("Error opening input gamma file.");
   }

   rootfilename=(char *)nulmemory(strlen(controlfilename)+1);
   strcpy(rootfilename,controlfilename);

   tempfile=Fopen("scratch.tmp","wt");

}

void close_files() {
   datafile=Fclose(datafile);
}

void initialize_globals() {
}

void alloc_record_structures() {
   int i;
   message("allocating record structures..");
   cellnum=(int *)getmemory((nsubjects+1)*sizeof(int));
   subjid =(char **)getmemory((nsubjects+1)*sizeof(char *));
   subjrec=(char **)getmemory((nsubjects+1)*sizeof(char *));
   for (i=0;i<=nsubjects;i++) {
      cellnum[i]=0;
      subjrec[i]=(char *)nulmemory(TJ+2);
      subjid[i]=(char *)nulmemory(keylen+1);
      subjid[i][keylen]=0;
   }
   howmuchleft();
}

void save_cellnum() {
   int i;
   for (i=1;i<=datafile->n;i++) {
      fprintf(cellfile->stream,"%-10.10s %05d",subjid[i],cellnum[i]);
      Fnewline(cellfile);
   }
   sprintf(tempstring,
      "Saved record keys and cell numbers to %s",cellfile->filename);
   log_message(tempstring);
}

void free_record_structures() {
   int i;
   message("releasing record structures..");
   for (i=0;i<=nsubjects;i++) {
      free(subjrec[i]);
      free(subjid[i]);
   }
   free(subjrec);
   subjrec=NULL;
   free(subjid);
   subjid=NULL;
   if (INITGAMMA-DATASET) {
      free(cellnum);
      cellnum=NULL;
   }
   howmuchleft();
}

void alloc_cell_structures() {
   int i;
   message("allocating packed data array..");
   cell     = (char **)getmemory((unsigned)(nsubjects+1)*sizeof(char *));
   cellfrq  = (FPPREC *)getmemory((unsigned)(nsubjects+1)*sizeof(FPPREC));
   for (i=0;i<=nsubjects;i++) {
      cell[i] = (char *)nulmemory(TJ+1);
      cellfrq[i] = 0.0;
   }
   howmuchleft();
}

void free_cell_structures() {
   int i;
   for (i=I+1;i<=nsubjects;i++)
      free(cell[i]);
}

void alloc_table_structures() {
   int i,j;

   message("allocating table structures..");

   levels   =(int * )nulmemory((TJ+1)*sizeof(int));
   baselevel=(int * )nulmemory((TJ+1)*sizeof(int));

   freqtable=(FPPREC **)nulmemory((TJ+1)*sizeof(FPPREC *));

   for (j=1;j<=TJ;j++)
      baselevel[j] = 32767;

   for (i=1;i<=I;i++)
      for (j=1;j<=TJ;j++) {
         baselevel[j] = min( baselevel[j], cell[i][j] );
         levels[j] = max( levels[j], cell[i][j] );
      }

   for (j=1;j<=TJ;j++) {
      levels[j] = 1 + levels[j] - baselevel[j];
      freqtable[j] = (FPPREC *)nulmemory((levels[j]+1)*sizeof(FPPREC));
   }
   howmuchleft();
}

void rescale_G(int i) {
   int k;
   double sumG;
   sumG=0.0;
   for (k=1;k<=K;k++) {
      if (G[i][k]<0.0) G[i][k]=0.0;
      sumG += G[i][k];
   }
   if (sumG < ZERO) {
      // sprintf(tempstring,"rescale_G: i=%d",i);
      // log_message(tempstring);
      sumG=(double)K;
      for (k=1;k<=K;k++)
         G[i][k] = 1.0;
   }
   for (k=1;k<=K;k++)
      G[i][k] = G[i][k]/sumG;
}

void rescale_P(int k,int j) {
   int l;
   double sumP;
   sumP=0.0;
   for (l=1;l<=levels[j];l++) {
      if (P[k][j][l]<=ZERO) P[k][j][l]=0.0;
      sumP += P[k][j][l];
   }
   if (sumP<ZERO) {
      // sprintf(tempstring,"rescale_P: k=%d j=%d levels[j]=%d",
      //    k,j,levels[j]);
      // log_message(tempstring);
      sumP=(double)levels[j];
      for (l=1;l<=levels[j];l++)
         P[k][j][l] = 1.0;
   } 
   for (l=1;l<=levels[j];l++)
      P[k][j][l] = P[k][j][l]/sumP;
}

void alloc_model_structures() {
   int i,j,k,l,pureI;
   FPPREC defG,defP;
   int randx;
   message("Allocating model structures..");
   if (P_gradient)
      log_message("Allocating LAMBDA gradient structures.");
   loglik=0.0;
   ploglik=0.0;
   dloglik=0.0;
   defG=(FPPREC)K;
   defG=1.0/defG;
   maxlevels=0;
   for (j=1;j<=TJ;j++)
      maxlevels=max(maxlevels,levels[j]);
   sprintf(tempstring,"maxlevels=%d",maxlevels);
   log_message(tempstring);

   G     = (FPPREC **)getmemory((I+1)*sizeof(FPPREC *));

   randomize();

   for (i=1; i<=I; i++) {

      G[i]     =(FPPREC *)nulmemory((K+1)*sizeof(FPPREC));

      switch (INITGAMMA) {
         case RANDOM:
            for (k=1;k<=K;k++) {
               randx=rand();
               G[i][k]=(FPPREC)randx;
            }
            break;
         case PURE1:
            pureI=1;
            for (j=1;pureI&&j<=J;j++)
               if (cell[i][j]-baselevel[j]) pureI=0;
            if (pureI) {
               G[i][1]=1.0;
               for (k=2;k<=K;k++)
                  G[i][k]=0.0;
            } else {
               for (k=1;k<=K;k++)
                  G[i][k]=1.0;
            }
            break;
         default:
            for (k=1;k<=K;k++)
               G[i][k]=1.0;
      }
      rescale_G(i);
   }

   P     = (FPPREC ***)getmemory((K+1)*sizeof(FPPREC **));
   new_P = (FPPREC ***)getmemory((K+1)*sizeof(FPPREC **));
   old_P = (FPPREC ***)getmemory((K+1)*sizeof(FPPREC **));
   fit_P = (int **)getmemory((K+1)*sizeof(int *));

   if (P_gradient) {
      dL_dP = (FPPREC ***)getmemory((K+1)*sizeof(FPPREC **));
      free_P= (char ***)getmemory((K+1)*sizeof(char **));
   }

   for (k=1; k<=K; k++) {
      P[k]     = (FPPREC **)getmemory((J+1)*sizeof(FPPREC *));
      new_P[k] = (FPPREC **)getmemory((J+1)*sizeof(FPPREC *));
      old_P[k] = (FPPREC **)getmemory((J+1)*sizeof(FPPREC *));
      fit_P[k] = (int *)nulmemory((J+1)*sizeof(int));

      if (P_gradient) {
         dL_dP[k] = (FPPREC **)getmemory((J+1)*sizeof(FPPREC *));
         free_P[k]= (char **)getmemory((J+1)*sizeof(char *));
      }

      for (j=1; j<=J; j++) {
         defP = levels[j];
         defP = 1.0/defP;
         P[k][j]     = (FPPREC *)nulmemory((maxlevels+1)*sizeof(FPPREC));
         new_P[k][j] = (FPPREC *)nulmemory((maxlevels+1)*sizeof(FPPREC));
         old_P[k][j] = (FPPREC *)nulmemory((maxlevels+1)*sizeof(FPPREC));
         if (P_gradient) {
            dL_dP[k][j] = (FPPREC *)nulmemory((maxlevels+1)*sizeof(FPPREC));
            free_P[k][j]= (char *)nulmemory(maxlevels+1);
         }
         fit_P[k][j] = 1;
      }
   }

   if (INITLAMBDA-DATASET) {
      for (j=1; j<=J; j++) {
         for (k=1;k<=K;k++) {
            for (l=1;l<=levels[j];l++) {
               switch (INITLAMBDA) {
                  case DEFAULT:
                     P[k][j][l]=1.0;
                     break;
                  case RANDOM:
                     randx=rand();
                     P[k][j][l] = (FPPREC)randx;
                     break;
                  case PURE1:
                     P[k][j][l]=1.0;
                     break;
               }
            }
         }
         if (INITLAMBDA==PURE1) {
            P[1][j][1]=1.0;
            for (l=2;l<=levels[j];l++) {
               P[1][j][l]=0.0;
            }
         }
         for (k=1;k<=K;k++)
            rescale_P(k,j);
      }
   }

   howmuchleft();
}

void tables() {
   int i,j;
   N = 0.0;
   for (i=1;i<=I;i++) {
      for (j=1;j<=TJ;j++)
         freqtable[j][1 + cell[i][j] - baselevel[j]] += cellfrq[i];
      N += cellfrq[i];
   }
}

void report_tables() {
   int i,j;
   FPPREC p;
   message("tabulating..");
   for (j=1;j<=TJ;j++) {
      log_message("");
      for (i=1;i<=levels[j];i++) {
         p=100.0*freqtable[j][i]/N;
         if (i-1)
            sprintf(tempstring,"            %4d.  %6.0lf   %6.1lf\%",i+baselevel[j]-1,freqtable[j][i],p);
         else
            sprintf(tempstring,"%10.10s  %4d.  %6.0lf   %6.1lf\%",varname[srcvarpos[j]],i+baselevel[j]-1,freqtable[j][i],p);
         log_message(tempstring);
      }
   }
   log_message("");
}

void check_dichotomous() {
   int j,i;
   DICHOTOMOUS=1;
   for (j=1;j<=J && DICHOTOMOUS;j++)
      if (levels[j]-2) DICHOTOMOUS=0;
   if (DICHOTOMOUS)
      log_message("Note: all internal variates are dichotomous.");
}

int endofrecord(char ch) {
   return (ch == 0 || ch == '\n' || ch == '\r') ? 1 : 0;
}

int delimiter(char ch) {
   if (endofrecord(ch))
      return 1;
   else
      return (ch == ' ' || ch == ',') ? 1 : 0;
}

void read_datafile() {
   int i,L,j,k,row,col;
   char ch,*rec,*token;
   int *x;

   rec=(char *)getmemory(MAXRECLEN+2);
   x=(int *)getmemory((nvars+1)*sizeof(int));
   token=(char *)getmemory(TOKENLEN+1);

   memset(token,0,TOKENLEN+1);

   message("reading input data from disk..");
   row=wherey();
   col=wherex();
   while (datafile->n<nsubjects) {
      memset(rec,0,MAXRECLEN+2);
      if (fgets(rec,MAXRECLEN,datafile->stream)==NULL) {
         datafile->eof=1;
         break;
      }
      datafile->n++;
      for (j=1;j<=nvars;j++)
         x[j]=(int)MISSING;

      if (NOID)
         ltoa(datafile->n,subjid[datafile->n],10);

      i=0;
      j=1;
      k=0;
      while (i<MAXRECLEN) {
         ch=rec[i++];
         if (delimiter(ch)) {
            if ( k ) {
               if (j+NOID==1)
                  strncpy(subjid[datafile->n],token,keylen);
               else
                  if (j<=nvars)
                     x[j-1+NOID]=atoi(token);
               j++;
               memset(token,0,TOKENLEN);
               k=0;
               if (endofrecord(ch))
                  break;
            }
         } else
            if (k<TOKENLEN)
               token[k++]=ch;
      }

      for (j=1;j<=TJ;j++) {
         i=srcvarpos[j];
         subjrec[datafile->n][j]=x[i];
      }

      if (datafile->n % 100 == 0) {
         gotoxy(col,row);
         sprintf(rec,"%d records.",datafile->n);
         cputs(rec);
      }

   }

   sprintf(rec,"%d records read from %s.",datafile->n,datafile->filename);
   message(rec);
   log_message(rec);
   log_message("");

   free(x);
   free(rec);
   free(token);
}

int read_lambdafile() {
   int Ok,k,j,l;
   double x;
   char buf[80];
   if (inlambdafile==NULL)
      return 1;
   log_message("reading pure type probabilities from disk..");
   Ok=1;
   for (j=1;j<=J;j++) {
      for (k=1;k<=K;k++) {
         for (l=1;l<=levels[j];l++) {
            if (fscanf(inlambdafile->stream,"%lf",&x)==1) {
               if (x<0.0) {
                  x=-x;
                  fit_P[k][j]=0;
               }
               P[k][j][l]=x;
            } else {
               P[k][j][l]=FMISSING;
            }
         }
         rescale_P(k,j);
      }
   }

   message("Initial pure type probabilities input from disk.");
   return Ok;
}

int read_gammafile() {
   int Ok,i,_i,k;
   double x;
   if (ingammafile==NULL)
      return 1;
   log_message("reading GOM scores from disk..");
   Ok=1;
   for (_i=1;_i<=datafile->n;_i++) {
      i=cellnum[_i];
      for (k=1;k<=K;k++) {
         if (fscanf(ingammafile->stream,"%lf",&x)==1) {
            G[i][k]=x;
            // sprintf(tempstring,"_i=%4d  i=%4d  k=%d  x=%10.4lf",_i,i,k,x);
            // log_message(tempstring);
         } else {
            G[i][k]=0.0;
         }
      }
      rescale_G(i);
   }
   message("Initial GOM scores input from disk.");
   return Ok;
}

void dump_datafile() {
   int i,j;
   for (i=1;i<=datafile->n;i++) {
      fprintf(tempfile->stream,"%s",subjid[i]);
      for (j=1;j<=J;j++)
         fprintf(tempfile->stream," %d",subjrec[i][j]);
      fprintf(tempfile->stream,"\n");
   }
}

void dump_G() {
   int i,k;
   char *s;
   s=(char *)nulmemory(20);
   for (i=1;i<=I;i++) {
      sprintf(tempstring,"cell #%4d.  G = ",i);
      for (k=1;k<=K;k++) {
         sprintf(s,"%10.4lf",G[i][k]);
         strcat(tempstring,s);
      }
      log_message(tempstring);
   }
   log_message("");
   free(s);
}

void dump_P() {
   int k,j,l,forced;
   char *s;
   s=(char *)nulmemory(20);
   /*
   for (k=1;k<=K;k++) {
      sprintf(tempstring,"Pure type probabilities for profile #%d.",k);
      log_message(tempstring);
      for (j=1;j<=J;j++) {
         sprintf(tempstring,"%10.10s, %d levels: ",varname[srcvarpos[j]],levels[j]);
         for (l=1;l<=levels[j];l++) {
            sprintf(s,"%10.4lf",P[k][j][l]);
            strcat(tempstring,s);
         }
         log_message(tempstring);
      }
      log_message("");
   }
   */
   log_message("");
   log_message("Pure type probabilities");
   log_message("-----------------------");
   log_message("");
   sprintf(tempstring,"         K = ");
   for (k=1;k<=K;k++) {
      sprintf(s,"%6d    ",k);
      strcat(tempstring,s);
   }
   log_message(tempstring);
   log_message("");
   forced=0;
   for (j=1;j<=J;j++) {
      for (l=1;l<=levels[j];l++) {
         if (l==1) {
            sprintf(tempstring,"%-10.10s %d ",varname[srcvarpos[j]],l+baselevel[j]-1);
         } else {
            sprintf(tempstring,"           %d ",l+baselevel[j]-1);
         }
         for (k=1; k<=K; k++) {
            if (fit_P[k][j]&&FITLAMBDA)
               sprintf(s,"%10.4lf",P[k][j][l]);
            else {
               forced=1;
               sprintf(s,"%9.4lf*",P[k][j][l]);
            }
            strcat(tempstring,s);
         }
         log_message(tempstring);
      }
      log_message("");
   }
   if (forced) {
      log_message("* held fixed");
      log_message("");
   }

   free(s);
}

int lessthan(char *a,char *b,int m) {
   int i,r;
   r=memcmp(a,b,m);
   return (r<0) ? 1 : 0;
}

void heapsort(int n,int ma,int mb,char *ra[],char *rb[]) {
   int l,j,ir,i;
   char *rra,*rrb;
   message("sorting..");
   rra=(char *)getmemory(ma+2);
   rrb=(char *)getmemory(mb+2);
   memset(rra,0,ma+2);
   memset(rrb,0,mb+2);
   l=(n >> 1)+1;
   ir=n;
   for (;;) {
      if (l > 1) {
         memcpy(rra,ra[--l],ma);
         memcpy(rrb,rb[l],mb);
      } else {
         memcpy(rra,ra[ir],ma);
         memcpy(rrb,rb[ir],mb);
         memcpy(ra[ir],ra[1],ma);
         memcpy(rb[ir],rb[1],mb);
         if (--ir == 1) {
            memcpy(ra[1],rra,ma);
            memcpy(rb[1],rrb,mb);
            free(rra);
            free(rrb);
            return;
         }
      }
      i=l;
      j=l << 1;
      while (j <= ir) {
         if (j < ir && lessthan(ra[j],ra[j+1],ma)) ++j;
         if (lessthan(rra,ra[j],ma)) {
            memcpy(ra[i],ra[j],ma);
            memcpy(rb[i],rb[j],mb);
            j += (i=j);
         }
         else j=ir+1;
      }
      memcpy(ra[i],rra,ma);
      memcpy(rb[i],rrb,mb);
   }
}

void heapsorti(int n,int ma,char *ra[],int rb[]) {
   int l,j,ir,i,rrb;
   char *rra;
   message("sorting..");
   rra=(char *)getmemory(ma+2);
   memset(rra,0,ma+2);
   rrb=0;
   l=(n >> 1)+1;
   ir=n;
   for (;;) {
      if (l > 1) {
         memcpy(rra,ra[--l],ma);
         rrb=rb[l];
      } else {
         memcpy(rra,ra[ir],ma);
         rrb=rb[ir];
         memcpy(ra[ir],ra[1],ma);
         rb[ir]=rb[1];
         if (--ir == 1) {
            memcpy(ra[1],rra,ma);
            rb[1]=rrb;
            free(rra);
            return;
         }
      }
      i=l;
      j=l << 1;
      while (j <= ir) {
         if (j < ir && lessthan(ra[j],ra[j+1],ma)) ++j;
         if (lessthan(rra,ra[j],ma)) {
            memcpy(ra[i],ra[j],ma);
            rb[i]=rb[j];
            j += (i=j);
         }
         else j=ir+1;
      }
      memcpy(ra[i],rra,ma);
      rb[i]=rrb;
   }
}

int count_cells(char *s[],int n,unsigned m) {
   int i,I;
   char *ppat;
   ppat=(char *)getmemory(m);
   memcpy(ppat,s[1],m);
   I=1;
   for (i=1;i<=n;i++)
      if (memcmp(s[i],ppat,m)) {
         I++;
         memcpy(ppat,s[i],m);
      }
   free(ppat);
   return I;
}

void pack_data() {
   int i,j;
   message("packing..");
   I   = count_cells(subjrec,datafile->n,TJ+1);
   memcpy(cell[1],subjrec[1],TJ+1);
   j=1;
   for (i=1;i<=datafile->n;i++) {
      if (memcmp(cell[j],subjrec[i],TJ+1))
         memcpy(cell[++j],subjrec[i],TJ+1);
      cellfrq[j]++;
      cellnum[i]=j;
   }
   sprintf(tempstring,"%d unique data patterns.",I);
   log_message(tempstring);
}

void dump_packedfile() {
   int i,j;
   message("saving packed file to disk..");
   for (i=1;i<=I;i++) {
      fprintf(tempfile->stream,"%6.0lf  ",cellfrq[i]);
      for (j=1;j<=TJ;j++)
         fprintf(tempfile->stream," %d",cell[i][j]);
      fprintf(tempfile->stream,"\n");
   }
}

void save_state() {
   int i,j,k,l;
   message("Saving scores and original data to disk..");
   for (i=1;i<=I;i++) {
      fprintf(gamfile->stream,"%05d %5.0lf   ",i,cellfrq[i]);
      for (j=1;j<=J;j++)
         fprintf(gamfile->stream,"%2d",cell[i][j]);
      for (k=1;k<=K;k++)
         fprintf(gamfile->stream,"%10.4lf",G[i][k]);
      Fnewline(gamfile);
   }

   for (j=1; j<=J; j++)
      for (k=1; k<=K; k++) {
         for (l=1; l<=levels[j]; l++)
            fprintf(lamfile->stream,"%10.4lf",P[k][j][l]);
         Fnewline(lamfile);
      }

   sprintf(tempstring,
      "Saved GOM scores and original data to %s",gamfile->filename);
   log_message(tempstring);

   sprintf(tempstring,
      "Saved pure type probabilities to %s",lamfile->filename);
   log_message(tempstring);
}

int isdelimiter(char ch) {
   if (ch <= ' ' || ch == ',' || ch == ';' || ch == '=')
      return 1;
   return 0;
}

#define INFILE 1
#define INLAMBDA 2
#define INGAMMA 3
#define INPUT 4
#define INTERNAL 5
#define EXTERNAL 6
#define PROFILES 7
#define RECORDS 8
#define GAMMAFIT 9
#define LAMBDAFIT 10
#define GRADIENTCMD 11

#define NCMDS 11

char CMDNAME[NCMDS+1][10]={
   "**NULL**","INFILE","INLAMBDA","INGAMMA","INPUT","INTERNAL","EXTERNAL",
      "PROFILES","RECORDS","GAMMAFIT","LAMBDAFIT","GRADIENT"};

int iscmd(char *s) {
   int i,n;
   for (i=1;i<=NCMDS;i++) {
      n=strlen(CMDNAME[i]);
      if (!strnicmp(s,CMDNAME[i],n))
         return i;
   }
   return 0;
}

int read_controlfile(char *s) {
   int n,cmd,p,i,b,j,ntokens,incomment,itemlen;
   char *rec,*item;
   char *token[MAXTOKENS];
   message("reading control file..");
   filename(&controlfilename,s);
   controlfile=Fopen(controlfilename,"rt");
   if (controlfile==NULL)
      crash("Error opening control file.");

   rec=(char *)getmemory(MAXRECLEN+2);
   ntokens=0; cmd=0; incomment=0;
   while (!controlfile->eof) {
      memset(rec,0,MAXRECLEN+2);
      if (fgets(rec,MAXRECLEN,controlfile->stream)==NULL) {
         controlfile->eof=1;
         break;
      }
      strip(rec,MAXRECLEN-1);
      controlfile->n++;
      log_message(rec);
      n=strlen(rec);
      p=0;
      while (p<=n) {

         if (!memcmp(rec+p,"/*",2)) {
            incomment=1;
            p+=2;
         }

         if (incomment) {
            if (!memcmp(rec+p,"*/",2)) {
               incomment=0;
               p+=2;
            } else {
               p++;
            }
         }

         if (incomment || p>n)
            continue;

         if (isdelimiter(rec[p])) {
            if (rec[p]==';') {

               switch (cmd) {

                  case INFILE :
                     filename(&datafilename,token[0]);
                     for (i=1;i<ntokens&&!NOID;i++)
                        if (!stricmp(token[i],"NOID"))
                           NOID=1;
                     break;

                  case INGAMMA :
                     if (!stricmp(token[0],"RANDOM"))
                        INITGAMMA=RANDOM;
                     else
                        if (!stricmp(token[0],"PURE1"))
                           INITGAMMA=PURE1;
                        else
                           if (!stricmp(token[0],"DEFAULT"))
                              INITGAMMA=DEFAULT;
                           else {
                              filename(&ingammafilename,token[0]);
                              INITGAMMA=DATASET;
                           }
                        for (i=1;i<ntokens;i++)
                           if (!stricmp(token[i],"NOFIT"))
                              FITGAMMA=0;
                     break;

                  case INLAMBDA :
                     if (!stricmp(token[0],"RANDOM"))
                        INITLAMBDA=RANDOM;
                     else
                        if (!stricmp(token[0],"DEFAULT"))
                           INITLAMBDA=DEFAULT;
                        else
                           if (!stricmp(token[0],"PURE1"))
                              INITLAMBDA=PURE1;
                           else {
                              filename(&inlambdafilename,token[0]);
                              INITLAMBDA=DATASET;
                           }
                        for (i=1;i<ntokens;i++)
                           if (!stricmp(token[i],"NOFIT"))
                              FITLAMBDA=0;
                     break;

                  case INPUT :
                     nvars=ntokens;
                     if (NOID) {
                        b=0;
                        varname[0]="RECNUM";
                     } else {
                        varname[0]=(char *)nulmemory(strlen(token[0]+1));
                        strcpy(varname[0],token[0]);
                        b=1;
                     }
                     for (i=1;i<=nvars;i++) {
                        varname[i]=(char *)nulmemory(strlen(token[i+b-1])+1);
                        strcpy(varname[i],token[i+b-1]);
                        internal[i]=0;
                        external[i]=0;
                     }
                     break;

                  case PROFILES :
                     K=atoi(token[0]);
                     break;

                  case INTERNAL :
                     J=0;
                     for (i=0;i<ntokens;i++) {
                        for (j=1;j<=nvars;j++) {
                           if (!stricmp(varname[j],token[i])) {
                              internal[j]=1;
                              J++;
                              break;
                           }
                        }
                        if (j>nvars) {
                           sprintf(tempstring,"INTERNAL variable %s not found.",
                              token[i]);
                           fatal_error(tempstring);
                        }
                     }
                     break;

                  case EXTERNAL :
                     JX=0;
                     for (i=0;i<ntokens;i++) {
                        for (j=1;j<=nvars;j++) {
                           if (!stricmp(varname[j],token[i])) {
                              JX++;
                              external[j]=1;
                              break;
                           }
                        }
                        if (j>nvars) {
                           sprintf(tempstring,"EXTERNAL variable %s not found.",
                              token[i]);
                           fatal_error(tempstring);
                        }
                     }
                     break;

                  case RECORDS :
                     nsubjects=atoi(token[0]);
                     break;

                  case GAMMAFIT:
                     if (!stricmp(token[0],"GRADIENT"))
                        G_algorithm = GRADIENT;
                     break;

                  case LAMBDAFIT:
                     if (!stricmp(token[0],"GRADIENT"))
                        P_algorithm = GRADIENT;
                     break;

                  case GRADIENTCMD:
                     P_algorithm = GRADIENT;
                     G_algorithm = GRADIENT;
                     break;

               }

               for (i=0;i<ntokens;i++)
                  free(token[i]);
               ntokens=0;
               cmd=0;
            }
            p++;
         } else {
            i=p;
            while (!isdelimiter(rec[p])) p++;
            itemlen=p-i;
            item=(char *)nulmemory(itemlen+1);
            memcpy(item,rec+i,itemlen);
            if (!cmd) {
               cmd=iscmd(item);
            } else {
               if (cmd) {
                  token[ntokens]=(char *)nulmemory(itemlen+1);
                  memcpy(token[ntokens],item,itemlen);
                  ntokens++;
               }
            }
            free(item);
         }
      }
   }

   log_message("");
   log_message("End of control file.");
   log_message("");

   if (P_algorithm)
      P_gradient=1;
   else
      P_gradient=0;

   if (G_algorithm)
      G_gradient=1;
   else
      G_gradient=0;

   for (i=0;i<ntokens;i++)
      free(token[i]);
   free(rec);
   TJ = J + JX;
   return 1;
}

void map_record() {
   int i,j,jx;
   j=0;
   jx=0;
   for (i=1;i<=nvars;i++)
      if ( internal[i] )
         srcvarpos[ ++j ] = i;
      else
         if ( external[i] )
            srcvarpos[ ++jx + J ] = i;
}

void control_summary() {
   int i,j;

   fullscreen();
   gotoxy((81-strlen(controlfilename))/2,1);
   textattr(WHITE);
   cputs(controlfilename);

   sprintf(tempstring,"\nControl file summary");
   log_message(tempstring);
   log_message("");
   sprintf(tempstring,"      Control file name : %s",controlfilename);
   log_message(tempstring);
   sprintf(tempstring,"             Case label : %s",varname[0]);
   log_message(tempstring);
   sprintf(tempstring,"        Input data file : %s",datafilename);
   log_message(tempstring);

   if (INITGAMMA==DATASET)
      sprintf(tempstring,"       Input gamma file : %s",ingammafilename);
   else
      sprintf(tempstring,"     Initial GOM scores : %s",INITSRCLABEL[INITGAMMA]);

   log_message(tempstring);

   sprintf(tempstring,"         Fit GOM scores : ");
   if (FITGAMMA)
      strcat(tempstring,algname[G_algorithm]);
   else
      strcat(tempstring,"No");
   log_message(tempstring);

   if (INITLAMBDA==DATASET)
      sprintf(tempstring,"      Input lambda file : %s",inlambdafilename);
   else
      sprintf(tempstring," Initial pure type prob : %s",INITSRCLABEL[INITLAMBDA]);

   log_message(tempstring);

   sprintf(tempstring,"     Fit Pure Type Prob : ");
   if (FITLAMBDA)
      strcat(tempstring,algname[P_algorithm]);
   else
      strcat(tempstring,"No");
   log_message(tempstring);

   sprintf(tempstring,"    #internal variables : %d",J);
   log_message(tempstring);
   sprintf(tempstring,"    #external variables : %d",JX);
   log_message(tempstring);

   sprintf(tempstring,"   max #records to read : %d",nsubjects);
   log_message(tempstring);


   sprintf(tempstring,"              #profiles : %d",K);
   log_message(tempstring);

   if (J) {
      log_message("");
      sprintf(tempstring,"%d internal variables:",J);
      log_message("");
      log_message(tempstring);
      for (j=1;j<=J;j++) {
         i = srcvarpos[j];
         sprintf(tempstring,"     %-10.10s",varname[i]);
         log_message(tempstring);
      }
   }

   if (JX) {
      log_message("");
      sprintf(tempstring,"  %d external variables:",JX);
      log_message("");
      log_message(tempstring);
      for (j=J+1;j<=TJ;j++) {
         i = srcvarpos[j];
         sprintf(tempstring,"     %-10.10s",varname[i]);
         log_message(tempstring);
      }
   }

   howmuchleft();
}

double cellgradient_G(int i,double *dL_dG,int *free_G) {
   int k,j,l,_k;
   double Pijl,norm;
   norm=0.0;
   for (k=1;k<=K;k++) {
      dL_dG[k]=-(double)J;
      for (j=1;j<=J;j++) {
         l = cell[i][j] + 1 - baselevel[j];
         Pijl=0.0;
         for (_k=1;_k<=K;_k++)
            Pijl = Pijl + G[i][_k]*P[_k][j][l];
         if (Pijl > ZERO)
            dL_dG[k] += P[k][j][l]/Pijl;
      }
      if ((G[i][k]<=ZERO && dL_dG[k] < 0.0) ||
         ((1.0-G[i][k])<=ZERO && dL_dG[k] > 0.0)) {
         free_G[k]=0;
      } else {
         free_G[k]=1;
         norm += dL_dG[k]*dL_dG[k];
      }
   }
   return norm;
}

double gradient_P() {
   int i,j,k,l,_k,_l,_j,stuck;
   double Pijl,Pij,Nij,Nijl,norm;
   norm=0.0;
   for (k=1;k<=K;k++) {
      for (j=1;j<=J;j++) {
         for (l=1;l<=levels[j];l++) {
            dL_dP[k][j][l]=0.0;
            for (i=1;i<=I;i++) {
               _l = cell[i][j] + 1 - baselevel[j];
               Pijl=0.0;
               for (_k=1;_k<=K;_k++)
                  Pijl += G[i][_k]*P[_k][j][l];
               if (l==_l) {
                  if (Pijl > ZERO)
                     dL_dP[k][j][l] +=
                        cellfrq[i]*G[i][k]*((1.0/Pijl)-1.0);
               } else {
                     dL_dP[k][j][l] -=
                        cellfrq[i]*G[i][k];
               }
            }
            
            if ((P[k][j][l]<=ZERO && dL_dP[k][j][l] < 0.0) ||
               ((1.0-P[k][j][l])<=ZERO && dL_dP[k][j][l] > 0.0)) {
               free_P[k][j][l]=0;
            } else {
               free_P[k][j][l]=1;
               norm += dL_dP[k][j][l]*dL_dP[k][j][l];
            }
            /*
            if ((P[k][j][l]<=ZERO) || ((1.0-P[k][j][l])<=ZERO)) {
               free_P[k][j][l]=0;
            } else {
               free_P[k][j][l]=1;
               norm += dL_dP[k][j][l]*dL_dP[k][j][l];
            }
            */
         }
      }
   }
   // getch();
   return norm;
}

double partlikelihood(int i) {
   double partlik,f_ijl;
   int j,k,l;
   partlik = 0.0;
   for (j=1; j<=J; j++) {
      l = cell[i][j] + 1 - baselevel[j];
      f_ijl=0.0;
      for (k=1;k<=K;k++) {
         f_ijl += G[i][k] * P[k][j][l];
      }
      if (f_ijl < ZERO) f_ijl=ZERO;
      if (f_ijl < REALBIG)
         partlik += log(f_ijl);
   }
   return partlik;
}

double loglikelihood() {
   int i,j,k,l;
   double sumlik,f_ijl,partlik;
   sumlik=0.0;
   ploglik=loglik;
   for (i=1; i<=I; i++) {
      partlik = 0.0;
      for (j=1; j<=J; j++) {
         l = cell[i][j] + 1 - baselevel[j];
         f_ijl=0.0;
         for (k=1;k<=K;k++) 
            f_ijl += G[i][k] * P[k][j][l];
         if (f_ijl < ZERO) f_ijl=ZERO;
         if (f_ijl < REALBIG)
            partlik += log(f_ijl);
      }
      sumlik += cellfrq[i]*partlik;
   }
   
   loglik=sumlik;
   return sumlik;
}

void restore_P() {
   int k,j,l;
   loglik=ploglik;
   for (k=1;k<=K;k++)
      for (j=1;j<=J;j++)
         for (l=1;l<=levels[j];l++)
            P[k][j][l] = old_P[k][j][l];
}

void loglik_message(char *s,double ploglik,int subiter) {
   sprintf(tempstring,"%10.10s  %3d.%03d %15.5lf %15.7lf",
      s,iter,subiter,loglik,fabs((loglik-ploglik)/ploglik));
   log_message(tempstring);
}

void fit_G_0() {
   int i,j,k,_k,l;
   double p_ijl;
   double *new_G;
   double *old_G;
   double *w;
   double partlik,newlik,startlik;
   int cnv,maxsubiter;

   w=(double *)getmemory((K+1)*sizeof(double));

   new_G = (double *)getmemory((K+1)*sizeof(double));
   old_G = (double *)getmemory((K+1)*sizeof(double));

   startlik=loglik;
   maxsubiter=0;

   for (i=1;i<=I;i++) {
      diddle();
      for (k=1;k<=K;k++)
         old_G[k]=G[i][k];
      partlik=partlikelihood(i);
      cnv=0;
      for (subiter=0; subiter<MAXITER2 && !cnv && !cancelled(); subiter++) {
         for (k=1;k<=K;k++) {
            w[k]=0.0;
            new_G[k]=0.0;
            for (j=1;j<=J;j++) {
               w[k]++;  /* looks stupid, will be used later */
               l = cell[i][j] + 1 - baselevel[j];
               p_ijl=0.0;
               for (_k=1;_k<=K;_k++)
                  p_ijl += G[i][_k]*P[_k][j][l];
               if (p_ijl > ZERO)
                  new_G[k] += G[i][k]*P[k][j][l]/p_ijl;
            }
         }
         for (k=1;k<=K;k++)
            G[i][k] = new_G[k] / w[k];
         rescale_G(i);
         newlik=partlikelihood(i);
         if (newlik < partlik) {
            for (k=1;k<=K;k++)
               G[i][k]=old_G[k];
            break;
         } else {
            if (partlik > ZERO) {
               if (fabs(newlik-partlik)/partlik < CTOL) {
                  cnv=1;
                  break;
               }
            }
         }
      }
      maxsubiter=max(maxsubiter,subiter);
   }

   loglikelihood();  // recalc loglik..
   loglik_message("fit_G_0",startlik,maxsubiter);
   
   log_message("");

   free(new_G);
   free(old_G);
   free(w);

}

void fit_P_0() {
   int k,j,l,i,_k;
   double g_ij, p_ijl, numer, denom, startlik, curlik;
   int cnv;

   startlik=loglik;
   cnv=0;

   for (subiter=0; subiter<MAXITER2 && !cnv && !cancelled(); subiter++) {
      
      for (k=1;k<=K;k++)
         for (j=1;j<=J;j++)
            for (l=1;l<=levels[j];l++)
               old_P[k][j][l] = P[k][j][l];
      
      for (k=1; k<=K; k++) {
         for (j=1; j<=J; j++) {
            if (fit_P[k][j]) {
               for (l=1; l<=levels[j]; l++) {
                  diddle();
                  numer=0.0;
                  denom=0.0;
                  for (i=1;i<=I;i++) {
                     p_ijl=0.0;
                     for (_k=1;_k<=K;_k++)
                        p_ijl += G[i][_k]*P[_k][j][l];
                     if (p_ijl > ZERO)
                        g_ij = G[i][k] * P[k][j][l] / p_ijl;
                     else
                        g_ij = 0.0;
                     denom += g_ij;
                     if ( l == ( 1 + cell[i][j] - baselevel[j] ) )
                        numer += g_ij;
                  }
                  if ( denom > ZERO )
                     new_P[k][j][l] = numer / denom;
                  else
                     new_P[k][j][l] = 0.0;
               }
            }
         }
      }

      for (k=1; k<=K; k++) {
         for (j=1; j<=J; j++) {
            if (fit_P[k][j]) {
               for (l=1; l<=levels[j]; l++) {
                  P[k][j][l] = new_P[k][j][l];
               }
               rescale_P(k,j);
            }
         }
      }

      curlik=loglikelihood();
      loglik_message("fit_P_0",startlik,subiter);
      if (curlik < startlik) {
         for (k=1;k<=K;k++)
            for (j=1;j<=J;j++)
               for (l=1;l<=levels[j];l++)
                  P[k][j][l] = old_P[k][j][l];
         loglik=startlik;
         break;
      } else {
         if (startlik > ZERO && fabs(curlik-startlik)/startlik < CTOL) {
            cnv=1;
            break;
         }
      }

      startlik=curlik;
   }
   log_message("");

}

// Congugate gradient (GRADIENT)

double MAXSTEP=1.0;
double MINSTEP=ZERO;
double GTOL=0.1;
int HALFSTEPS=25;

void fit_G_1() {
   double partlik,oldlik,loglikchange,holdstep,startlik,
      old_norm_G,f0,f1,fprime0,s,norm,oldnorm,beta,stepsize,x;
   int i,j,k,l,h,converged,restarted,restarts,nbound,halfstep,somefree,
      maxsubiter;
   long subiters;
   double *oldG, *dL_dG, *bestG;
   double bestlik;
   int *free_G;

   oldG  =(double *)nulmemory((K+1)*sizeof(double));
   dL_dG =(double *)nulmemory((K+1)*sizeof(double));
   bestG =(double *)nulmemory((K+1)*sizeof(double));
   free_G=(int    *)nulmemory((K+1)*sizeof(int   ));

   startlik=loglik;
   maxsubiter=0;
   subiters=0L;

   for ( i=1; i<=I && !cancelled(); i++ ) {

      converged=0;
      diddle();
      for (subiter=0;
//         subiter<MAXITER2 && converged==0 && !cancelled();
         converged==0 && !cancelled();
         subiter++) {

         norm  = cellgradient_G(i,dL_dG,free_G);
         if (norm <= ZERO) {
            converged=1;
            break;
         }

         somefree=0;
         for (k=1;k<=K;k++) {
            oldG[k]=G[i][k];
            bestG[k]=G[i][k];
            if (free_G[k]) somefree++;
         }

         if (!somefree) break;

         f0=partlikelihood(i);
         stepsize=MAXSTEP;
         bestlik=f0;

         for (halfstep=0;
            halfstep < HALFSTEPS && !cancelled() && stepsize > MINSTEP;
            halfstep++) {

            // new estimates
            for (k=1;k<=K;k++)
                G[i][k]=oldG[k] + stepsize * dL_dG[k];
            rescale_G(i);

            f1=partlikelihood(i);

            // best fit this iteration..
            if (f1 > bestlik) {
               bestlik=f1;
               for (k=1;k<=K;k++)
                  bestG[k]=G[i][k];
            } else {
               if (bestlik > f0+ZERO && bestlik > f1+ZERO)
                  break;
            }

            stepsize=stepsize/2.0;

         } // halfstep..

         // rebuild best fit
         f1=bestlik;
         for (k=1;k<=K;k++)
            G[i][k]=bestG[k];

         if (fabs(f0) > ZERO &&
            fabs((f1-f0)/f0) < CTOL)
            converged=1;

         if (fabs(f0) > ZERO &&
            fabs((f1-f0)/f0) < ITOL &&
            subiter > MAXITER2)
            break;

         subiters++;

      } // subiter..

   } // i..

   // recalc loglik
   loglikelihood();

   subiters=subiters/(long)I;
   maxsubiter=(int)subiters;

   loglik_message("fit_G_1",startlik,maxsubiter);
   log_message("");

   free(oldG );
   free(dL_dG);
   free(bestG);
   free(free_G);

}

void fit_P_1() {
   double curlik,oldlik,
      f0,f1,fprime0,s,norm,oldnorm,beta,stepsize,x,startlik;
   int i,j,k,l,converged,restarted,restarts,halfstep,somefree;
   double ***oldP, ***bestP;
   double bestlik;

   oldP  = (double ***)getmemory((K+1)*sizeof(double **));
   bestP = (double ***)getmemory((K+1)*sizeof(double **));

   for (k=1;k<=K;k++) {
      oldP[k]  = (double **)getmemory((J+1)*sizeof(double *));
      bestP[k] = (double **)getmemory((J+1)*sizeof(double *));
      for (j=1; j<=J; j++) {
         oldP[k][j]  = (double *)nulmemory((maxlevels+1)*sizeof(double));
         bestP[k][j] = (double *)nulmemory((maxlevels+1)*sizeof(double));
      }
   }

   converged=0;
   for (subiter=0;
//      subiter<MAXITER2 && converged==0 && !cancelled();
      converged==0 && !cancelled();
      subiter++) {

      norm  = gradient_P();

      if (norm <= ZERO) {
         converged=1;
         break;
      }

      somefree=0;

      for (k=1;k<=K;k++)
         for (j=1;j<=J;j++)
            for (l=1;l<=levels[j];l++) {
               oldP[k][j][l] = P[k][j][l];
               bestP[k][j][l] = P[k][j][l];
               if (free_P[k][j][l]) somefree++;
            }

      if (!somefree)
         break;

      stepsize=MAXSTEP;
      f0=loglikelihood();
      bestlik=f0;

      for (halfstep=0;
         halfstep < HALFSTEPS && converged==0 && !cancelled();
         halfstep++) {

         diddle();

         // new estimates
         for (k=1;k<=K;k++)
            for (j=1;j<=J;j++)
               if (fit_P[k][j]) {
                  for (l=1;l<=levels[j];l++)
                     P[k][j][l]=oldP[k][j][l] + stepsize * dL_dP[k][j][l];
                  rescale_P(k,j);
               }

         // best fit this iteration..

         f1=loglikelihood();

         if (f1 > bestlik) {
            bestlik=f1;
            for (k=1;k<=K;k++)
               for (j=1;j<=J;j++)
                  for (l=1;l<=levels[j];l++)
                      bestP[k][j][l]=P[k][j][l];
         } else {
            if (bestlik > f0+ZERO && bestlik > f1+ZERO)
               break;
         }

         stepsize=stepsize/2.0;
         
      } // halfstep..

      // best fit..
      f1 = bestlik;
      loglik = bestlik;
      for (k=1;k<=K;k++)
         for (j=1;j<=J;j++)
            for (l=1;l<=levels[j];l++)
               P[k][j][l]=bestP[k][j][l];

      // log_message(tempstring);

      if (fabs(f0) > ZERO &&
         fabs((f1-f0)/f0) < CTOL)
         converged=1;
         
      if (fabs(f0) > ZERO &&
         fabs((f1-f0)/f0) < ITOL &&
         subiter >= MAXITER2)
         break;


      loglik_message("fit_P_1",f0,subiter);

   } // subiter..

   for (k=1;k<=K;k++) {
      for (j=1; j<=J; j++) {
         free(oldP[k][j]);
         free(bestP[k][j]);
      }
      free(oldP[k]);
      free(bestP[k]);
   }
   free(oldP);
   free(bestP);

   log_message("");
}

int model() {
   int converged;
   double norm,x;
   FPPREC lastlik,dlik;
   int i,k,j,l;
   void (*fit_P)(),(*fit_G)();
   time_t time_started;
   CANCELLED=0;

   if (P_algorithm==GRADIENT)
      fit_P = fit_P_1;
   else
      fit_P = fit_P_0;
   
   if (G_algorithm==GRADIENT)
      fit_G = fit_G_1;
   else
      fit_G = fit_G_0;
   
   lastlik=loglik;

   time_started=time(NULL);

   fullscreen();
   gotoxy(1,25);
   textattr(WHITE);
   clreol();
   cputs("press ESC to cancel");

   message("Fitting GOM model..");
   log_message("");
   sprintf(tempstring,"Algorithm for GOM scores is %s.",algname[G_algorithm]);
   log_message(tempstring);
   sprintf(tempstring,"Algorithm for pure type probabilities is %s.",algname[P_algorithm]);
   log_message(tempstring);
   log_message("");
   loglikelihood();
   sprintf(tempstring,"Initial log-likelihood is %-10.4lf.",loglik);
   log_message(tempstring);
   log_message("");
   for (iter=0; iter<MAXITER && !cancelled(); iter++) {

      if (cancelled())
         break;

      dlik=loglik-lastlik;
      lastlik=loglik;

      if (iter) {
         if (fabs(dlik/loglik) < CTOL)
            break;
      }

      if (FITGAMMA)
         if ( INITGAMMA != DATASET || iter > 0 )
            (*fit_G)();

      if (FITLAMBDA)
         (*fit_P)();

   }

   if (CANCELLED) {
      log_message("");
      log_message("*cancelled*");
   }

   loglikelihood();
   sprintf(tempstring,"Final log-likelihood is %-10.4lf.",loglik);
   log_message(tempstring);
   log_message("");

   /*
   if (GRADIENT) {
      norm=0.0;
      for (i=1;i<=I;i++) {
         norm += cellgradient_G(i);
         for (k=1;k<=K;k++) {
            sprintf(tempstring,"i=%3d k=%2d G=%20.10lf dL=%15.5lf",
               i,k,G[i][k],dL_dG[i][k]);
            log_message(tempstring);
         }
      }

      sprintf(tempstring,"norm is %15.5lf",norm);
      log_message(tempstring);

      gradient_P();
      for (k=1;k<=K;k++)
         for (j=1;j<=J;j++)
            for (l=1;l<=levels[j];l++) {
               sprintf(tempstring,"k=%d j=%2d l=%d  P=%15.8lf dL=%15.8lf",
                  k,j,l,P[k][j][l],dL_dP[k][j][l]);
               log_message(tempstring);
            }
   }
   */

   fullscreen();
   gotoxy(1,25);
   textattr(WHITE);
   clreol();
   i=(int)(time(NULL)-time_started);
   i=i/60;
   sprintf(tempstring,"Gom3 consumed %d minutes for this problem.",i);
   log_message(tempstring);
   activate_window(CURWIN);

   return converged;
}

void wrapup_msg() {
   sprintf(tempstring,"Minimum amount of free RAM was %ld bytes.",minfreeram);
   log_message(tempstring);
   log_message("");
}

void main(int argc, char *argv[]) {
   int i;

   if (argc<2) {
      crash("\nNo control file specified\n");
      exit(8);
   }

   open_log(argv[1]);

   draw_mainscreen();

   log_message(version);
   log_message("");

   if (!read_controlfile(argv[1]))
      crash("Error reading control file");

   validate_specs();
   if (ERRORS)
      crash("Fatal error(s) found in control file.");

   map_record();

   control_summary();

   open_files();

   open_model_files(NULL);

   alloc_cell_structures();

   alloc_record_structures();

   read_datafile();

   heapsort(datafile->n,TJ+1,keylen,subjrec,subjid);

   pack_data();

   save_cellnum();

   free_record_structures();

   free_cell_structures();

   alloc_table_structures();

   tables();

   report_tables();

   check_dichotomous();

   alloc_model_structures();

   read_lambdafile();

   read_gammafile();

   dump_P();

   model();

   save_state();

   dump_P();

   close_files();

   wrapup_msg();

   Exit(0);
}

